#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

//#include "detri2.h"

//using  namespace detri2;

//==============================================================================
#define LINEBUF 1024

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// findnextnumber()   Find the next field of a number string.                //
//                                                                           //
// Jumps past the current field by searching for whitespace or a comma, then //
// jumps past the whitespace or the comma to find the next field that looks  //
// like a number.                                                            //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

char* findnextnumber(char *string)
{
  char *result;

  result = string;
  // Skip the current field.  Stop upon reaching whitespace or a comma.
  while ((*result != '\0') && (*result != '#') && (*result != ' ') && 
         (*result != '\t') && (*result != ',')) {
    result++;
  }
  // Now skip the whitespace and anything else that doesn't look like a
  //   number, a comment, or the end of a line. 
  while ((*result != '\0') && (*result != '#')
         && (*result != '.') && (*result != '+') && (*result != '-')
         && ((*result < '0') || (*result > '9'))) {
    result++;
  }
  // Check for a comment (prefixed with `#').
  if (*result == '#') {
    *result = '\0';
  }
  return result;
}

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// findnextfield()   Find the next field of a string.                        //
//                                                                           //
// Jumps past the current field by searching for whitespace or a comma, then //
// jumps past the whitespace or the comma to find the next field.            //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

char* findnextfield(char *string)
{
  char *result;

  result = string;
  // Skip the current field.  Stop upon reaching whitespace or a comma.
  while ((*result != '\0') &&
         (*result != ' ') &&
         (*result != '\t') &&
         (*result != ',') &&
         (*result != ';')) {
    result++;
  }
  // Now skip the whitespace or the comma, stop at anything else that looks
  //   like a character, or the end of a line. 
  while ((*result == ' ') ||
         (*result == '\t') ||
         (*result == ',') ||
         (*result == ';') ||
         (*result == '{') ||
         (*result == '}')) {
    result++;
  }
  return result;
}

//==============================================================================

class CMapVertex {
 public:
  int idx;
  double x, y, z;
  int farther;
  double r, g, b;
  double u, v;

  void init() {
    idx = 0;
    x = y = z = 0.0;
    farther = 0;
    r = g = b = 0.0;
    u = v = 0.0;
  }
};

class CMapFace {
 public:
  int idx;
  int v1, v2, v3;
  int farther;

  void init() {
    idx = 0;
    v1 = v2 = v3 = 0;
    farther = 0;
  }
};

int get_xy_uv_mesh(int argc, char* argv[])
{
  if (argc != 2) {
    printf("Usage: get_xyz_uv dump_bamg.uv.m\n");
    return 0;
  }

  // Read the file once, get the nnode, nface.
  char buf[LINEBUF]; //, new_buf[LINEBUF];
  //char delim[] = " ,\t";
  char *pstr;

  int nnode, nface;
  int ncount, fcount;
  int idx;

  nnode = 0; // 3422;
  nface = 0; // 6674;
  ncount = fcount = 0; // Count the number vertices, faces.
  
  char filename1[LINEBUF];
  strcpy(filename1, argv[1]);

  FILE *infile = NULL;
  infile = fopen(filename1, "r");
  if (infile == NULL) {
    printf("Usage: get_xyz_uv dump_bamg.uv.m\n");
    return 0;
  }
  printf("Reading %s\n", filename1);

  // The bounding box.
  double xmin, xmax, ymin, ymax, zmin, zmax;
  xmin = ymin = zmin =  1.e+30;
  xmax = ymax = zmax = -1.e+30;
  double x, y, z;

  double umin, umax, vmin, vmax;
  umin = vmin =  1.e+30;
  umax = vmax = -1.e+30;
  double u, v;

  while (fgets(buf, LINEBUF, infile) != NULL) {
    //sprintf(new_buf, "%3d %s", count, buf);
    pstr = buf;
    if (strstr(pstr, "Vertex")) {
      pstr = findnextnumber(pstr);
      idx = atoi(pstr);
      if (idx > nnode) nnode = idx;
      pstr = findnextnumber(pstr);
      x = atof(pstr);
      pstr = findnextnumber(pstr);
      y = atof(pstr);
      pstr = findnextnumber(pstr);
      z = atof(pstr);
      xmin = x < xmin ? x : xmin;
      xmax = x > xmax ? x : xmax;
      ymin = y < ymin ? y : ymin;
      ymax = y > ymax ? y : ymax;
      zmin = z < zmin ? z : zmin;
      zmax = z > zmax ? z : zmax;
      // Search for keyword "uv"
      pstr = findnextfield(pstr);
      while (*pstr != '\0') {
        if (!strncmp(pstr, "uv=", 3)) {
          while (*pstr != '(') pstr++;
          pstr++; // skip '(';
          u = atof(pstr);
          pstr = findnextnumber(pstr);
          v = atof(pstr);
          umin = u < umin ? u : umin;
          umax = u > umax ? u : umax;
          vmin = v < vmin ? v : vmin;
          vmax = v > vmax ? v : vmax;
        }
        pstr = findnextfield(pstr);
      }
      
      ncount++;
    } else if (strstr(pstr, "Face")) {
      //fputs(buf, outfile);
      pstr = findnextnumber(pstr);
      idx = atoi(pstr);
      if (idx > nface) nface = idx;
      fcount++;
    } else if (strstr(pstr, "Edge")) {
      //fputs(buf, outfile);
    }
  }
  printf("Number of nodes = %d, ncount=%d\n", nnode, ncount);
  printf("Number of faces = %d, fcount=%d\n", nface, fcount);
  fclose(infile);

  if (nnode != ncount) {
    printf("!! Warning:  wrong number of vertices.\n");
  }
  if (nface != fcount) {
    printf("!! Warning:  wrong number of faces. use fcount=%d\n", fcount);
  }

  printf("Boundong box [x,y,z]:\n");
  printf("  x: [%g, %g]\n", xmin, xmax);
  printf("  y: [%g, %g]\n", ymin, ymax);
  printf("  z: [%g, %g]\n", zmin, zmax);

  printf("Boundong box [u,v]:\n");
  printf("  u: [%g, %g]\n", umin, umax);
  printf("  v: [%g, %g]\n", vmin, vmax);

  infile = fopen(filename1, "r");
  printf("Reading %s\n", filename1);

  // Read the list of vertices.
  CMapVertex *vlist = new CMapVertex[nnode];

  int i;
  // Initialise all vertices.
  for (i = 0; i < nnode; i++) {
    CMapVertex *v = &(vlist[i]);
    v->init();
    v->idx = i+1;
    v->x = 0.5 * (xmin + xmax);
    v->y = 0.5 * (ymin + ymax);
    v->z = 0.5 * (zmin + zmax);
    v->u = 0.5 * (umin + umax);
    v->v = 0.5 * (vmin + vmax);
  }

  // Read exactly ncount Vertices
  for (i = 0; i < ncount; i++) {
    fgets(buf, LINEBUF, infile);
    pstr = buf;
    // Read Vertex idx x y z
    assert(!strncmp(pstr, "Vertex", 6));
    pstr = findnextnumber(pstr);
    idx = atoi(pstr);

    CMapVertex *v = &(vlist[idx-1]);
    //v->init();
    assert(v->idx == idx);
    pstr = findnextnumber(pstr);
    v->x = atof(pstr);
    pstr = findnextnumber(pstr);
    v->y = atof(pstr);
    pstr = findnextnumber(pstr);
    v->z = atof(pstr);

    // Read father, rgb, uv
    pstr = findnextfield(pstr);
    while (*pstr != '\0') {
      if (!strncmp(pstr, "father=", 7)) {
        while (*pstr != '(') pstr++;
        pstr++; // skip '(';
        v->farther = atoi(pstr);
      } else if (!strncmp(pstr, "rgb=", 4)) {
        //pstr = findnextnumber(pstr);
        while (*pstr != '(') pstr++;
        pstr++; // skip '(';
        v->r = atof(pstr);
        pstr = findnextnumber(pstr);
        v->g = atof(pstr);
        pstr = findnextnumber(pstr);
        v->b = atof(pstr);
      } else if (!strncmp(pstr, "uv=", 3)) {
        //pstr = findnextnumber(pstr);
        while (*pstr != '(') pstr++;
        pstr++; // skip '(';
        v->u = atof(pstr);
        pstr = findnextnumber(pstr);
        v->v = atof(pstr);
      }
      pstr = findnextfield(pstr);
    }
  } // Vertex

  CMapFace *flist = new CMapFace[fcount];

  for (i = 0; i < fcount; i++) {
    fgets(buf, LINEBUF, infile);
    pstr = buf;
    // Read Face idx v1 v2 v3
    assert(!strncmp(pstr, "Face", 4));
    CMapFace *f = &flist[i];
    pstr = findnextnumber(pstr);
    idx = atoi(pstr);
    f->idx = idx;
    pstr = findnextnumber(pstr);
    f->v1 = atoi(pstr);
    pstr = findnextnumber(pstr);
    f->v2 = atoi(pstr);
    pstr = findnextnumber(pstr);
    f->v3 = atoi(pstr);

    // Read father, rgb, uv
    pstr = findnextfield(pstr);
    while (*pstr != '\0') {
      if (!strncmp(pstr, "father=", 7)) {
        //pstr = findnextnumber(pstr);
        while (*pstr != '(') pstr++;
        pstr++; // skip '(';
        f->farther = atoi(pstr);
      }
      pstr = findnextfield(pstr);
    }
  }

  //=========================================
  FILE* xyzsmesh = fopen("output_xyz.smesh", "w");
  printf("Writing %s\n", "output_xyz.smesh");
  fprintf(xyzsmesh, "%d 3 0\n", nnode);
  for (i = 0; i < nnode; i++) {
    CMapVertex *v = &(vlist[i]);
    fprintf(xyzsmesh, "%d %g %g %g\n", v->idx, v->x, v->y, v->z);
  }
  fprintf(xyzsmesh, "%d 0\n", fcount);
  for (i = 0; i < fcount; i++) {
    CMapFace *f = &(flist[i]);
    fprintf(xyzsmesh, "3 %d %d %d\n", f->v1, f->v2, f->v3);
  }
  fprintf(xyzsmesh, "0\n");
  fclose(xyzsmesh);

  FILE *uvsmesh  = fopen("output_uv.smesh", "w");
  printf("Writing %s\n", "output_uv.smesh");
  fprintf(uvsmesh, "%d 3 0\n", nnode);
  for (i = 0; i < nnode; i++) {
    CMapVertex *v = &(vlist[i]);
    fprintf(uvsmesh, "%d %g %g 0\n", v->idx, v->u, v->v);
  }
  fprintf(uvsmesh, "%d 0\n", fcount);
  for (i = 0; i < fcount; i++) {
    CMapFace *f = &(flist[i]);
    fprintf(uvsmesh, "3 %d %d %d\n", f->v1, f->v2, f->v3);
  }
  fprintf(uvsmesh, "0\n");
  fclose(uvsmesh);

  delete [] vlist;
  delete [] flist;
  return 1;
}

//==============================================================================

int main(int argc, char* argv[])
{
  return get_xy_uv_mesh(argc, argv);
}
