#include "hde.h"
#include <math.h>

//==============================================================================
// Gaussian
// see an example from BossenHeckbert96-anisoMesh.pdf, page 8.

double fun_gaussian(const detri2::Vertex * vert);
double funX_gaussian(const detri2::Vertex * vert);
double funY_gaussian(const detri2::Vertex * vert);

// Hessian
double funXX_gaussian(const detri2::Vertex * vert);
double funXY_gaussian(const detri2::Vertex * vert);
double funYY_gaussian(const detri2::Vertex * vert);

//==============================================================================
// lennard
// f2(x,y)=tanh(60x)-tanh(60(x-y)-30)

double fun_lennard(const detri2::Vertex * vert);
double funX_lennard(const detri2::Vertex * vert);
double funY_lennard(const detri2::Vertex * vert);

double funXX_lennard(const detri2::Vertex * vert);
double funXY_lennard(const detri2::Vertex * vert);
double funYY_lennard(const detri2::Vertex * vert);

//==============================================================================
// f3(x,y)=sin(5 * pow(x - 0.2, 3)*(y*y - y + 1))

double fun_alauzet(const detri2::Vertex * vert);
double funX_alauzet(const detri2::Vertex * vert);
double funY_alauzet(const detri2::Vertex * vert);

double funXX_alauzet(const detri2::Vertex * vert);
double funXY_alauzet(const detri2::Vertex * vert);
double funYY_alauzet(const detri2::Vertex * vert);
//==============================================================================
// This function is an example in Huang's paper (Huang05-AnisoMetric.pdf example 4.2, page 23)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net
// tanh(-100*(y-0.5-0.4*sin(2pix))^2)

double fun_Huang(const detri2::Vertex * vert);
double funX_Huang(const detri2::Vertex * vert);
double funY_Huang(const detri2::Vertex * vert);

//==============================================================================
// Du, Q.; Faber, V. & Gunzburger, M.
// Centroidal Voronoi Tessellations: Applications and Algorithms SIAM Rev.,
// Society for Industrial and Applied Mathematics, 1999, 41, 637-676
// Page 670, Figure 7.5

//double alpha = 40;

double fun_du(const detri2::Vertex * vert);
double funX_du(const detri2::Vertex * vert);
double funY_du(const detri2::Vertex * vert);

//==============================================================================
// This function is an example of the BAMG program (at manual.pdf page 12)
// Use the following website to get the derivatives of f_x and f_y.
// https://www.derivative-calculator.net

//f5=(10x^3+y^3)+atan2(0.001,sin(5y)-2x)
//    +(10y^3+x^3+atan(0.01,sin(5x)-2y))

double fun_bamg(const detri2::Vertex * vert);
double funX_bamg(const detri2::Vertex * vert);
double funY_bamg(const detri2::Vertex * vert);

//==============================================================================

//double x0 = 0.25;
//double x1 = -0.25;

double fun_Gaussian2(const detri2::Vertex * vert);
double funX_Gaussian2(const detri2::Vertex * vert);
double funY_Gaussian2(const detri2::Vertex * vert);

//==============================================================================
// Rosenbrock function
// see https://en.wikipedia.org/wiki/Rosenbrock_function
// f = (a-x)^2+b*(y-x^2)^2
// fx = -4*b*x*(y-x^2)-2*(a-x)
// fy = 2*b*(y-x^2);

//double coef_a = 1.;
//double coef_b = 100.;

double fun_Rosenbrock(const detri2::Vertex * vert);
double funX_Rosenbrock(const detri2::Vertex * vert);
double funY_Rosenbrock(const detri2::Vertex * vert);

//==============================================================================

double fun_thanks(const detri2::Vertex * vert);
double funX_thanks(const detri2::Vertex * vert);
double funY_thanks(const detri2::Vertex * vert);

//==============================================================================
//f1(x,y)=tanh(2(x-y)-1)

//==============================================================================
// The quadratic function, z = x^+y^2

double fun_Delaunay(const detri2::Vertex * vert);
double funX_Delaunay(const detri2::Vertex * vert);
double funY_Delaunay(const detri2::Vertex * vert);

//==============================================================================
// A saddle surface is a smooth surface containing one or more saddle points.
// The hyperbolic paraboloid (hyperboloid of one sheet): z = x^2 - y^2 or z = xy

double fun_Saddle(const detri2::Vertex * vert);
double funX_Saddle(const detri2::Vertex * vert);
double funY_Saddle(const detri2::Vertex * vert);

//==============================================================================
