#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "detri2.h"

using  namespace detri2;

//==============================================================================

// If 'pt' is given, only check if S is encroached by pt.
// Otherwise, check the apex of S.
// In Delaunay refinement case, 'pt' must be given.
int Triangulation::check_segment(Vertex *pt, TriEdge &S, arraypool *encsegs)
{
  assert(S.is_segment());
  Vertex *e1 = S.org();
  Vertex *e2 = S.dest();

  if (pt != NULL) {
    assert(pt != tr_infvrt);
    double v1[2], v2[2];
    double costh = 0;
    v1[0] = e1->crd[0] - pt->crd[0];
    v1[1] = e1->crd[1] - pt->crd[1];
    v2[0] = e2->crd[0] - pt->crd[0];
    v2[1] = e2->crd[1] - pt->crd[1];
    costh = v1[0] * v2[0] + v1[1] * v2[1];
    if (costh < 0) {
      Triang *pE = (Triang *) encsegs->alloc();
      pE->init();
      pE->vrt[0] = e1;
      pE->vrt[1] = e2;
      pE->vrt[2] = NULL; // Don't save a rejected vertex.
      return 1;
    }
    return 0;
  }

  for (int i = 0; i < 2; i++) {
    Vertex *chkpt = S.apex();
    if (chkpt != tr_infvrt) {
      double v1[2], v2[2];
      double costh = 0;
      v1[0] = e1->crd[0] - chkpt->crd[0];
      v1[1] = e1->crd[1] - chkpt->crd[1];
      v2[0] = e2->crd[0] - chkpt->crd[0];
      v2[1] = e2->crd[1] - chkpt->crd[1];
      costh = v1[0] * v2[0] + v1[1] * v2[1];
      if (costh < 0) {
        Triang *pE = (Triang *) encsegs->alloc();
        pE->init();
        pE->vrt[0] = e1;
        pE->vrt[1] = e2;
        pE->vrt[2] = chkpt; // Save the encroached point.
        return 1;
      }
    }
    S = S.esym();
  }

  if (op_metric == METRIC_ISO) { // =1
    // Check if this segment satisifies mesh size.
    TriEdge E; E.tri = S.get_segment();
    printf("Check segment %d, %d -- val=%g\n", e1->idx, e2->idx, E.tri->val);
    if (E.tri->val > 0) {
      double len = get_distance(e1, e2);
      printf(" len = %g\n", len);
      if (len > E.tri->val) {
        Triang *pE = (Triang *) encsegs->alloc();
        pE->init();
        pE->vrt[0] = e1;
        pE->vrt[1] = e2;
        pE->vrt[2] = NULL; // Don't save a rejected vertex.
        return 1; // split this segment.
      }
    }
  }

  return 0;
}

//==============================================================================

int Triangulation::split_enc_segment(TriEdge &S, Vertex *encpt, arraypool *fqueue,
                                 arraypool *encsegs, arraypool *enctris)
{
  //if (op_db_verbose > 2) {
    printf("    Split encroached segment [%d, %d]\n", S.org()->idx, S.dest()->idx);
  //}

  if (!S.is_segment()) {
    // report a bug
    //strcpy(io_outfilename, "/Users/si/tmp/dump-bug");
    //save_triangulation();
    assert(0);
  }

  Triang *seg = S.get_segment();
  assert(seg != NULL);
  int stag = seg->tag;
  REAL val = seg->val;
  printf("    val = %g\n", val);

  Vertex *newvrt = (Vertex *) tr_steiners->alloc();
  newvrt->init();

  // Default split it at the middle.
  Vertex *e1 = S.org();
  Vertex *e2 = S.dest();
  newvrt->crd[0] = 0.5 * (e1->crd[0] + e2->crd[0]);
  newvrt->crd[1] = 0.5 * (e1->crd[1] + e2->crd[1]);

  if ((encpt != NULL)) {
    assert(!encpt->is_deleted()); // It must be an existing vertex.
    if ((encpt->on_bd.tri != NULL) && (seg != encpt->on_bd.tri)) {
      // Check if these two segments shara a common vertex.
      if ((seg->vrt[0] == encpt->on_bd.tri->vrt[0]) ||
          (seg->vrt[0] == encpt->on_bd.tri->vrt[1])) {
        // Shared at seg->vrt[0]
        REAL split = get_distance(seg->vrt[0], encpt) /
                     get_distance(seg->vrt[0], seg->vrt[1]);
        newvrt->crd[0] = seg->vrt[0]->crd[0] +
          split * (seg->vrt[1]->crd[0] - seg->vrt[0]->crd[0]);
        newvrt->crd[1] = seg->vrt[0]->crd[1] +
          split * (seg->vrt[1]->crd[1] - seg->vrt[0]->crd[1]);
      } else if ((seg->vrt[1] == encpt->on_bd.tri->vrt[0]) ||
                 (seg->vrt[1] == encpt->on_bd.tri->vrt[1])) {
        // Shared at seg->vrt[1]
        REAL split = get_distance(seg->vrt[1], encpt) /
                     get_distance(seg->vrt[0], seg->vrt[1]);
        newvrt->crd[0] = seg->vrt[1]->crd[0] +
          split * (seg->vrt[0]->crd[0] - seg->vrt[1]->crd[0]);
        newvrt->crd[1] = seg->vrt[1]->crd[1] +
          split * (seg->vrt[0]->crd[1] - seg->vrt[1]->crd[1]);
      }
    }
  }

  newvrt->crd[2] = newvrt->crd[0] * newvrt->crd[0] + newvrt->crd[1] * newvrt->crd[1];
  newvrt->idx = io_firstindex + (ct_in_vrts + tr_steiners->objects - 1);
  newvrt->tag = stag; //seg->tag;
  newvrt->typ = STEINERVERTEX;

  //int stag = seg->tag; // done above
  remove_segment(seg); // Remove this segment.
  //assert(seg->is_deleted());
  //printf("   nnnnnnnnnnn\n");
  //if (check_mesh(0,0) > 0) {
  //  assert(0);
  //}

  // Insert the new point (split the segment).
  int fflag = FLIP_24;
  TriEdge tt[4];
  tt[0] = S; // [e1, e2, c]
  tt[2] = tt[0].esym(); // [e2, e1, d]
  flip(tt, &newvrt, fflag, fqueue);

  S = tt[1].enext(); // [e1, newpt, c]
  assert(S.org() == e1);
  assert(S.dest() == newvrt);
  insert_segment(S, stag, val, NULL);
  if (encsegs != NULL) {
    check_segment(NULL, S, encsegs);
  }

  S = tt[0].eprev(); // [newpt, e2, d]
  assert(S.org() == newvrt);
  assert(S.dest() == e2);
  insert_segment(S, stag, val, NULL);
  if (encsegs != NULL) {
    check_segment(NULL, S, encsegs);
  }

  // For Delaunay refinement, we fix the segmment Steiner vertex
  //   to prevent to remove them, this will cause endless loop.
  assert(!newvrt->is_fixed());
  newvrt->set_fix();

  // Debug split_enc_segment() only
  //if (check_mesh(0,0) > 0) {
  //  assert(0);
  //}

  lawson_flip(newvrt, 0, fqueue); // hullflag = 0

  if (enctris != NULL) {
    // Add the triangles in the star of newvrt into list.
    enq_vertex_star(newvrt, enctris);
  }

  return 1;
}

//==============================================================================

int Triangulation::repair_encsegments(arraypool *encsegs, arraypool *enctris)
{
  arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
  TriEdge E;

  while (encsegs->objects > 0) {
    for (int i = 0; i < encsegs->used_items; i++) {
      Triang *pE = (Triang *) encsegs->get(i);
      if (pE->is_deleted()) continue;
      if (get_edge(pE->vrt[0], pE->vrt[1], E)) {
        split_enc_segment(E, pE->vrt[2], fqueue, encsegs, enctris);
      }
      pE->set_deleted();
      encsegs->dealloc(pE);
    }
  }

  encsegs->clean();

  delete fqueue;
  return 1;
}

//==============================================================================

void Triangulation::enq_triangle(Triang* tri, REAL cct[3], arraypool* enctris)
{
  Triang *paryEle = (Triang *) enctris->alloc();
  paryEle->init(); // Initialize
  paryEle->nei[0].tri = tri;
  for (int j = 0; j < 3; j++) {
    paryEle->vrt[j] = tri->vrt[j];
  }
  for (int j = 0; j < 3; j++) {
    paryEle->cct[j] = cct[j];
  }
}

void Triangulation::enq_vertex_star(Vertex *pt, arraypool* enctris)
{
  TriEdge E = pt->adj;
  do {
    check_triangle(E.tri, enctris);
    E = E.eprev_esym();
  } while (E.tri != pt->adj.tri);
}

int Triangulation::check_triangle(Triang *tri, arraypool* enctris)
{
  // Skip a hull and exterior triangle.
  if (tri->is_hulltri() || 
      tri->is_exterior()) return 0;

  // torg  -> tri->vrt[0]->crd
  // tdest -> tri->vrt[1]->crd
  // tapex -> tri->vrt[2]->crd
  REAL xdo = tri->vrt[1]->crd[0] - tri->vrt[0]->crd[0];
  REAL ydo = tri->vrt[1]->crd[1] - tri->vrt[0]->crd[1];
  REAL xao = tri->vrt[2]->crd[0] - tri->vrt[0]->crd[0];
  REAL yao = tri->vrt[2]->crd[1] - tri->vrt[0]->crd[1];
  REAL dodist = xdo * xdo + ydo * ydo;
  REAL aodist = xao * xao + yao * yao;
  REAL dadist = (tri->vrt[1]->crd[0] - tri->vrt[2]->crd[0]) *
                (tri->vrt[1]->crd[0] - tri->vrt[2]->crd[0]) +
                (tri->vrt[1]->crd[1] - tri->vrt[2]->crd[1]) *
                (tri->vrt[1]->crd[1] - tri->vrt[2]->crd[1]);
  REAL denominator = 0.5 / (xdo * yao - xao * ydo); // area.
  REAL dx = (yao * dodist - ydo * aodist) * denominator;
  REAL dy = (xdo * aodist - xao * dodist) * denominator;

  double ccent[3];
  ccent[0] = tri->vrt[0]->crd[0] + dx;
  ccent[1] = tri->vrt[0]->crd[1] + dy;

  if (op_minlen > 0.0) { // -L#
    REAL radius = sqrt(dx * dx + dy * dy);
    REAL diff = radius - op_minlen;
    if (diff < 0.0) {
      return 0; // Reject to split this triangle
    }
  }

  REAL area = 0.5 * fabs(xdo * yao - xao * ydo);
  if (op_maxarea > 0) { // -a#
    //REAL area = 0.5 * (xdo * yao - xao * ydo);
    if (area > op_maxarea) {
      enq_triangle(tri, ccent, enctris);
      return 1; // Split triangle.
    }
  }

  if (op_metric == METRIC_ISO) { // -m
    //printf("  check tri %d,%d,%d, tag(%d), val=%g, area=%g\n", tri->vrt[0]->idx,
    //       tri->vrt[1]->idx, tri->vrt[2]->idx, tri->tag, tri->val, area);
    if ((tri->val > 0) && (area > tri->val)) {
      enq_triangle(tri, ccent, enctris);
      return 1; // Split triangle.
    }

    if (OMT_domain != NULL) {
      // A background mesh is given.
      /*
      // Check mesh size at the triangle (.area file)
      Vertex pt;
      pt.crd[0] = (tri->vrt[0]->crd[0]+tri->vrt[1]->crd[0]+tri->vrt[2]->crd[0]) / 3.;
      pt.crd[1] = (tri->vrt[0]->crd[1]+tri->vrt[1]->crd[1]+tri->vrt[2]->crd[1]) / 3.;
      TriEdge E; E.tri=NULL;
      int loc = OMT_domain->locate_point(&pt, E, 1, 0); // round = 1, encflag = 0
      if (loc == LOC_IN_TRI) { // in triangle
        // E.tri is the background triangle.
        printf("  E.tri->val=%g, area=%g\n", E.tri->val, area);
        if ((E.tri->val > 0) && (area > E.tri->val)) {
          enq_triangle(tri, ccent, enctris);
          return 1; // Split triangle.
        }
      }
      */
      // Check the mesh size at the three vertices of this triangle (.mtr file).
      REAL radius = sqrt(dx * dx + dy * dy);
      double val = 0.;
      for (int i = 0; i < 3; i++) {
        val = 0.0;
        //printf("aaaaaa tri->vrt[i]=%d\n", tri->vrt[i]->idx);
        if (get_vertex_metric(tri->vrt[i]) > 0) {
          //printf("bbbbbb tri->vrt[i]=%d, val=%g\n", tri->vrt[i]->idx, tri->vrt[i]->val);
          val = tri->vrt[i]->val;
          if (val > 0) {
            //printf("  tri->vrt[i]=%d, val=%g\n", tri->vrt[i]->idx, tri->vrt[i]->val);
            REAL diff = radius - val;
            REAL tol = fabs(diff) / val;
            if ((tol > 0.2) && (diff > 0)) {
              enq_triangle(tri, ccent, enctris);
              return 1; // Split triangle.
            }
          }
        }
      }
    } // if (OMT_domain != NULL)
  }

  if (op_minangle > 0) {
    TriEdge E;
    E.tri = tri;
    REAL minlen2;
    if (dodist < aodist) {
      if (dodist < dadist) { 
        E.ver = 2; // org->dest
        minlen2 = dodist;
      } else {
        E.ver = 0; // dest->apex
        minlen2 = dadist;
      }
    } else {
      if (aodist < dadist) {
        E.ver = 1; // apex->org
        minlen2 = aodist;
      } else {
        E.ver = 0; // dest->apex
        minlen2 = dadist;
      }
    }
    //if (minedgelen > 0) {
    //  // TO DO: Check if we need to delete a point.
    //}
  
    REAL radius2 = dx * dx + dy * dy;
    REAL ratio2 = radius2 / minlen2;
    if (ratio2 > op_maxratio2) {
      // Check if there are two segments.
      if (E.enext().is_segment() && E.eprev().is_segment()) {
        // Do not split a triangle containing a small angle.
        // Miller, Pav, and Walkington's adaptive version.
        return 0;
      }
      enq_triangle(tri, ccent, enctris);
      return 1;
    }
  } // if (op_maxratio2 > 0)

  return 0;
}

int Triangulation::get_triangle(Vertex *e1, Vertex *e2, Vertex *e3, TriEdge& E)
{
  if ((e1 == NULL) || (e2 == NULL) || (e3 == NULL)) {
    assert(0);
    return 0;
  }

  if (e1->is_deleted() || (e1->typ == UNUSEDVERTEX) ||
      e2->is_deleted() || (e2->typ == UNUSEDVERTEX) ||
      e3->is_deleted() || (e3->typ == UNUSEDVERTEX)) {
    return 0;
  }
  if (get_edge(e1, e2, E)) {
    if (E.apex() == e3) return 1;
    E = E.esym();
    return E.apex() == e3;
  }
  return 0;
}

//==============================================================================

// Assume all triangles to be checked are in chktris, and are infected.
int Triangulation::repair_triangles(arraypool *enctris)
{
  TriEdge E, S, tt[4];
  int loc, i;

  int est_size = tr_segs->objects * 4;
  int log2objperblk = 1;
  while (est_size >>= 1) log2objperblk++;
  arraypool *encsegs = new arraypool(sizeof(Triang), log2objperblk);

  arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);

  while (enctris->objects > 0) {
    for (i = 0; i < enctris->used_items; i++) {
      Triang *qtri = (Triang *) enctris->get(i);
      if (qtri->is_deleted()) continue;
      if (get_triangle(qtri->vrt[0], qtri->vrt[1], qtri->vrt[2], E)) {
        // Try to split it.
        Vertex *newvrt = (Vertex *) tr_steiners->alloc();
        newvrt->init();
        newvrt->crd[0] = qtri->cct[0];
        newvrt->crd[1] = qtri->cct[1];
        newvrt->crd[2] = newvrt->crd[0] * newvrt->crd[0] + newvrt->crd[1] * newvrt->crd[1];
        newvrt->idx = io_firstindex + (ct_in_vrts + tr_steiners->objects - 1);
        newvrt->typ = STEINERVERTEX;
        //loc = locate_point(newvrt, E, 0, 1); // encflag = 1
        loc = locate_point(newvrt, E, 1); // encflag = 1
        if (loc == LOC_ENC_SEG) {
          // Try to cross a segment. Split it.
          Triang *pE = (Triang *) encsegs->alloc();
          pE->init();
          pE->vrt[0] = E.org();
          pE->vrt[1] = E.dest();
          //pE->vrt[2] = NULL;
          newvrt->set_deleted();
          tr_steiners->dealloc(newvrt);
        } else {
          // Insert the vertex.
          assert(!E.tri->is_hulltri());
          tt[0] = E;
          if (loc == LOC_IN_TRI) {
            int fflag = FLIP_13;
            flip(tt, &newvrt, fflag, fqueue);
          } else if (loc == LOC_ON_EDGE) {
            int fflag = FLIP_24;
            flip(tt, &newvrt, fflag, fqueue);
          } else {
            assert(0); // on vertex? Must be a bug
          }
          lawson_flip(newvrt, 0, fqueue);
          // Check if this vertex encroach other segments.
          E = newvrt->adj;
          do {
            S = E.enext(); // its oppsite edge.
            if (S.is_segment()) {
              check_segment(newvrt, S, encsegs);
            }
            E = E.eprev_esym();
          } while (E.tri != newvrt->adj.tri);
          if (encsegs->objects > 0) {
            remove_point(newvrt, fqueue);
            lawson_flip(NULL, 0, fqueue);
          }
        }
        if (encsegs->objects > 0) {
          repair_encsegments(encsegs, enctris);
        } else {
          enq_vertex_star(newvrt, enctris);
        }
      } // if (get_triangle)
      qtri->set_deleted();
      enctris->dealloc(qtri);
    } // i
  } // while

  delete fqueue;
  delete encsegs;
  //delete enctris;
  return 1;
}

//==============================================================================

// This function implements Ruppert's Delaunay refinement algorithm [1995].
// It uses Si and Gartner [2005]'s CDT edge splitting algorithm to protect
// encroached segments with sharp angles. Miller, Pav, and Walkington [2003]
// showed that Rupper's algorithm terminates as long as the sharp angle is
// no less than 30 degree (theoretically 26.53 degree).

int Triangulation::delaunay_refinement()
{
  //clock_t tv[4]; // Timing informations (defined in time.h)
  //REAL cps = (REAL) CLOCKS_PER_SEC;

  //tv[0] = clock();

  if (tr_steiners == NULL) {
    tr_steiners = new arraypool(sizeof(Vertex), 13); // 2^13 = 8192
  }

  if ((op_metric == METRIC_ISO) || (op_minlen > 0.0)) { // -m or -L#
    coarsen_mesh();
  }

  //mesh_phase = 4;

  if (!op_no_gabriel) { // no -G  Gabriel property is default.
    // Put all encroached segments into queue.
    arraypool *encsegs = new arraypool(sizeof(Triang), 10);

    TriEdge E;
    for (int i = 0; i < tr_tris->objects; i++) {
      E.tri = (Triang *) tr_tris->get(i);
      if (!E.tri->is_deleted()) {
        if (!E.tri->is_hulltri()) { 
          for (E.ver = 0; E.ver < 3; E.ver++) {
            if (E.is_segment()) {
              check_segment(NULL, E, encsegs);
            }
          }
        }
      }
    }

    repair_encsegments(encsegs, NULL);

    delete encsegs;
  }

  if (op_quality || (op_metric == METRIC_ISO) || (op_maxarea > 0.0)) { // -q or -a#
    if (op_db_verbose) {
      printf("Adding Steiner points to enforce quality.\n");
    }
    // Calculate the squre of the maximum radius-edge ratio.
    // r/d = 1/(2*sin(theta_min));
    op_maxratio2 = 0.5 / sin(PI * op_minangle / 180.0);
    op_maxratio2 *= op_maxratio2;

    int est_size = tr_tris->objects * 2;
    int log2objperblk = 0;
    while (est_size >>= 1) log2objperblk++;
    arraypool *enctris = new arraypool(sizeof(Triang), log2objperblk);

    // Put all triangles into list
    for (int i = 0; i < tr_tris->used_items; i++) {
      Triang *tri = (Triang *) tr_tris->get(i);
      if (tri->is_deleted()) continue;
      check_triangle(tri, enctris);
    }

    repair_triangles(enctris);

    delete enctris;
  }

  //tv[1] = clock();
  //if (db_verbose) {
  //  printf("  Delaunay refinement seconds:  %g\n", ((REAL)(tv[1]-tv[0])) / cps);
  //}

  //mesh_phase = 0;

  return 1;
}
