#ifndef DETRI2QT_H
#define DETRI2QT_H

#include "detri2/detri2.h"

// Initialize an triangulation with 4 vertices on bbox.
void detri2_init_with_bbox(detri2::Triangulation **pTr, double xmin, double xmax, double ymin, double ymax);

void detri2_projpt2edge(detri2::Vertex *p, detri2::Vertex* e1, detri2::Vertex* e2, detri2::Vertex* prj);

void detri2_get_vertex_typ(detri2::Vertex *p, char strtyp[64]);

// Locte the point (x, y) in the triangulation.
//   Return 1 (in triangle), 2 (on edge), 3 (on vertex).
//   E returns the triangle, moreover, E.org() is the nearest vertex to (x,y).
//   dist returns the distance from its neartest vertex.
int detri2_locate_point(detri2::Triangulation* Tr, double x, double y, detri2::TriEdge &E,
                        double &dist, int randflag = 1);

// Insert the point (x,y) into the triangulation.
//   If E.tri != NULL, or loc == LOC_IN_OUTSIDE, then E and loc are returned from the locate_point();
//   Otherwise, detri2_locate_point() will be called.
//   The new vertex is returned by E.org().
//   [Comment]: This function does not insert a point on line segment!
int detri2_insert_point(detri2::Triangulation* Tr, double x, double y, detri2::TriEdge &E,
                        int& loc, int randflag = 1);

// Locate the neartest edge to the point (x,y).
//   It should always find an edge (Even (x,y) lies outside).
//   If loc != LOC_IN_OUTSIDE, the returned triangle E must contain (x,y), moreover,
//   If loc == LOC_IN_TRIANGLE, ang returns the value at vertex E.org(), E.dest(), and pt(x,y).
//     It is used to check if we should split the edge (segment) by (x,y).
int detri2_locate_edge(detri2::Triangulation* Tr, double x, double y, detri2::TriEdge &E);
//int detri2_get_segment(detri2::TriEdge &E, detri2::Triang **pseg);

// Insert an edge e1->e2 into the triangulation (geometry).
// This function calls recover_edge().
int detri2_insert_segment(detri2::Triangulation* Tr,
                          detri2::Vertex *e1, detri2::Vertex *e2, int tag, REAL val,
                          detri2::arraypool *fqueue);

// Check if the edge [e1,e2] is intersected by any segment or vertex.
//#define INTERSECT_NONE           0   // initial
//#define INTERSECT_SHARE_EDGE     1   // A edge exists
//#define INTERSECT_CONFLICT       2   // A segment already exists
//#define INTERSECT_VERTEX         3   // cutting by a vertex
//#define INTERSECT_SEGMENT        4   // intersecting another segment
//int detri2_detect_intersection(detri2::Triangulation* Tr, detri2::Vertex *e1, detri2::Vertex *e2,
//                               detri2::TriEdge &S);
//int detri2_insert_segment_intersect(detri2::Triangulation* Tr,
//                                    detri2::Vertex *e1, detri2::Vertex *e2, int tag,
//                                    detri2::TriEdge &S);

// Removing a segment at vertex, update the vertex-to-segment ring.
//   'v' must be an endpoint of a segment => v->typ is either RIDGEVERTEX or v->is_fixed().
//int detri2_remove_segment_at_vertex(detri2::Vertex *v, detri2::Triang *seg);

// Move the vertex (*ppt) of the triangulation to the new position (x,y).
//   Return the vertex at (x,y), its weight remain unchanged.
int detri2_move_vertex(detri2::Triangulation* Tr, detri2::Vertex **ppt, double x, double y);

// Get all triangles in a given subdomain.
int detri2_get_trilist(detri2::Triang *tri, detri2::arraypool *tri_list);
int detri2_is_exterior_trilist(detri2::arraypool *tri_list);
int detri2_is_tri_in_trilist(detri2::arraypool *tri_list, detri2::Triang* testtri);

int detri2_get_subdomain(detri2::Triangulation* Tr, detri2::arraypool *tri_list, detri2::Vertex **sdm_vertex);
int detri2_add_subdomain(detri2::Triangulation** pTr, detri2::arraypool *tri_list, int tag, double maxarea);

// Enlarge the bounding box to include the point (x,y). It does the following:
//   - delete the current triangulation in Tr,
//   - enlarge the couurent bbox to include the point (x,y),
//   - create the inital triangulation of the bbox,
//   - insert all Steiner points of Tr.
int detri2_enlarge_bbox(detri2::Triangulation** pTr, double x, double y);
int detri2_enlarge_bbox(detri2::Triangulation** pTr,
                        double xmin, double xmax, double ymin, double ymax);

// Call rbox to generate a set of random points.
int detri2_rbox(detri2::Triangulation** pTr, int dim, int numpoints);
int detri2_rbox(detri2::Triangulation** pTr, int argc, char *argv[]);

// Add a rectangle with (x, y) as its lower-left corner and the given width and height.
int detri2_add_rectangle(detri2::Triangulation** pGeom,
                         double ctx, double cty, double width, double height, int etag,
                         bool transflag, double rcx, double rcy, double angle, double dx, double dy);
int detri2_add_circle(detri2::Triangulation** pGeom,
                      double ctx, double cty, double radius, double ang1, double ang2, int np, int etag,
                      bool transflag, double rcx, double rcy, double angle, double dx, double dy);
int detri2_add_geometry(detri2::Triangulation** pGeom,
                        int np, double *ptlist, int ne, int *elist, int *etaglist);

// Scale the vertices into the range [bbxmin, bbxmax]x[bbymin, bbymax]
int detri2_scale_points(detri2::Triangulation* Geom, double bbxmin, double bbxmax, double bbymin, double bbymax);

//void detri2_Hilbert_sort();

// Sweepline incremental construction
void detri2_sweepline_sort(detri2::Vertex **ptlist, int ptnum, double vx, double vy);

// Create the (nearest or furthest) weighted Delaunay triangulation.
int detri2_Delaunay(detri2::Triangulation* Geom, detri2::Triangulation** pTr,
                    int furthest, int weighted, int cdt);

// Generate mesh from Geom.
int detri2_generate(detri2::Triangulation* Geom, detri2::Triangulation** pMesh, int weighted, int cdt);

// Load detri2 from files.
int detri2_reconstruct(detri2::Triangulation** pGeom, detri2::Triangulation** pTr, const char* filename);

// Mesh adaptation through HDE
int detri2_HDE_mesh_adapt(detri2::Triangulation** MeshIO,
                          double s1,
                          double s2,
                          double min_len,
                          double min_angle,
                          double size_iter,
                          double qual_iter,
                          double use_intpoints,
                          double use_smooth,
                          double smooth_iter,
                          double smooth_alpha,
                          int verbose,
                          int test_fun);

// Mesh smoothing -- Huang's MMPDE approach
//int detri2_mesh_smooth_mmpde(detri2::Triangulation** pMesh, double tf, double abstol, double dt);

// drawing functions
void detri2_draw_point(detri2::Vertex* v, int r, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_line(detri2::Vertex* e1, detri2::Vertex* e2, void* Painter, double Cx, double Cy, double Sxy);

void detri2_draw_bbox(double _xmin, double _xmax, double _ymin, double _ymax,
                      void* Painter, double Cx, double Cy, double Sxy);
//void detri2_draw_geometry(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
//void detri2_draw_segments(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_convex_hull(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_Triangulation(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
//void detri2_draw_subdomains(detri2::Triangulation* Geom, void* Painter, double Cx, double Cy, double Sxy);

//void detri2_draw_weights(detri2::Triangulation* Geom, detri2::Triangulation* tmpTr, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_orthocircles(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_voronoi(detri2::Triangulation* Tr, detri2::Triangulation* OMT,
                         void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_voronoi_mass_centers(detri2::Triangulation* Tr, detri2::Triangulation* OMT,
                                      bool show_power_cell, int center_size,
                                      void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_voronoi_cell(detri2::Triangulation* Tr, detri2::Triangulation* OMT, detri2::Vertex *mesh_vertex,
                              bool show_mass_center, int center_size, void* Painter, double Cx, double Cy, double Sxy);
void detri2_draw_voronoi_cell_debug(detri2::Triangulation* Tr, detri2::Triangulation* OMT, detri2::Vertex *mesh_vertex,
                              void* Painter, double Cx, double Cy, double Sxy);

// Show Laplacian centers of vertices, which are arithmetic average of the neighboring points.
void detri2_draw_laplacian_centers(detri2::Triangulation* Tr, int center_size,
                                   void* Painter, double Cx, double Cy, double Sxy);

//void detri2_draw_vertex_labels(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);
//void detri2_draw_segment_labels(detri2::Triangulation* Tr, void* Painter, double Cx, double Cy, double Sxy);

void detri2_draw_hilbert_init(int n);
void detri2_draw_Hilbert_curve(double bxmin, double bxmax, double bymin, double bymax,
                               int e, int d, int order, int depth,
                               double *EntryPt, double *ExitPt, int *hindex,
                               void* Painter, double Cx, double Cy, double Sxy);

#endif // DETRI2QT_H
