#ifndef SCRIBBLEAREA_H
#define SCRIBBLEAREA_H

#include <QColor>
#include <QImage>
#include <QWidget>
#include <QMenuBar>
#include <QStatusBar>
#include <QToolBar>

#include "detri2/detri2.h"
#include "widgets/newfilepanel.h"
#include "widgets/meshoptionpanel.h"
#include "widgets/hdeoptionpanel.h"
#include "widgets/sortoptionpanel.h"
#include "widgets/vertexpanel.h"
#include "widgets/segmentpanel.h"
#include "widgets/subdomainpanel.h"
#include "widgets/imageview.h"
#include "widgets/preferdialog.h"

class ScribbleArea : public QWidget
{
    Q_OBJECT

public:
    ScribbleArea(QToolBar *toolbar_top, QToolBar *toolbar_left,
                 QStatusBar *statusbar, QMenuBar *menubar,
                 QWidget *parent = 0);

    detri2::Triangulation *_geom; // the geometry
    detri2::Triangulation *_tria; // the triangulation

    QImage _canvas[2]; // [0] foreground, [1] background (copy).
    int    _object_valid;

    QImage *_background_img;

    double _cx, _cy, _scale_xy; // affine transformation from (x,y) to (X,Y)
                                // (x,y) are input coordinates,
                                // (X,Y) are screen coordinates, such that
                                // X =  _scale_xy * x + cx, and
                                // Y = -_scale_xy * y + cy,

    double  _mouse_sens;
    int     _mouse_last_x;  /**< last mouse x position in window */
    int     _mouse_last_y;  /**< last mouse y position in window */
    int     _mouse_down;    /**< mouse button pressed flag */
    int     _mouse_down_move;

    // Mouse edit options
    //bool    _trans_mode;
    bool    _insert_mode;
    //bool    _insert_vertex;
    //bool    _insert_segment;
    //bool    _insert_rect;
    //bool    _insert_arc;
    bool    _select_mode;
    //bool    _assign_vertex;
    //bool    _assign_weight;
    //bool    _assign_segment;
    //bool    _assign_subdomain;
    //bool    _view_power_cell;
    //bool    _view_triangle_info;
    //bool    _view_triangle_orthocircle;
    //bool    _view_triangle_orthocenter;

    //bool    _zoom_selected_region;
    int     _select_x, _select_y, _select_width, _select_height;

    // View options
    //bool    _show_geom_vertices;
    //bool    _show_geom_edges;
    //bool    _show_geom_subdomains;
    //bool    _show_geom_bbox;
    //bool    _show_geom_tris;
    //bool    _show_weights;
    //bool    _show_convex_hull;
    //bool    _show_triangles;
    //bool    _show_ortho_circle;
    //bool    _show_voronoi;
    //bool    _show_vertex_labels;
    //bool    _show_segment_labels;

    detri2::Vertex*     _active_vertex;
    detri2::Triang*     _active_segment;
    detri2::arraypool*  _active_trilist;
    detri2::Vertex*     _active_sdm_vertex;

    newfilePanel*       _newfilepabel;
    meshoptionPanel*    _meshoptionPanel;
    hdeoptionPanel*     _hdeoptionPanel;
    sortOptionPanel*    _sortoptionpanel;
    vertexPanel*        _vertexPanel;
    segmentPanel*       _segmentPanel;
    subdomainPanel*     _subdomainPanel;
    ImageView*          _imgView;
    PreferDialog*       _preferdialog;

    QToolBar*           _toolbar_top;
    QToolBar*           _toolbar_left;

    QStatusBar*         _statusbar;
    QLabel*             m_statusLeft;
    QLabel*             m_statusMiddle;
    QLabel*             m_statusRight;

    QMenuBar*           _menubar;

    // View options
    int                 _draw_vertex_size; // 4
    int                 _draw_segment_width; // 2
    int                 _draw_edge_width; // 1
    int                 _draw_dual_edge_width; // 1
    int                 _draw_ccircle_width; // 1
    QColor              _color_canvas_background; // Qt::white
    QColor              _color_vertex;
    QColor              _color_segment; // Qt::blue
    QColor              _color_edge;
    QColor              _color_dual_edge;
    QColor              _color_ccircle; //
    QColor              _color_selected;
    QFont               _font_bbox_label;
    QFont               _font_vertex_label;
    QFont               _font_segment_label;
    QFont               _font_subdomain_label;
    QFont               _font_meshvertex_label;

    // Debug options
    int                 _enable_debug;
    QString             _file_debug;

    QString             _file_last_used_path;

public slots:
    void clean_scribble();
    void reset_scribble();
    void reset_preference();

//private slots:
    void newFile();
    void openFile();
    void addFile();
    void saveFile();
    void closeFile();
    void open_preferDialog();
    void on_user_exit();

    // Edit
    void edit_mouse_act();
    void edit_attributes_act();
    void edit_delete_act();
    void edit_scale_act();

    // View
    void view_update_act();
    void view_trans_act(int i);
    void view_zoom_act(int i);
    void view_reset();

    void detri2_options_act();
    void generate_mesh_act();
    void generate_mesh_ani_act();
    void weighted_Delaunay_act();
    void furthest_Delaunay_act();
    void detri2_HDE_option_act();
    void detri2_HDE_act();
    void detri2_Sort_option_act();
    void detri2_Sort_act();
    void detri2_mesh_smooth_act();
    void clean_mesh_act();

private:
    // file menu
    QAction*    newAct;
    QAction*    openAct;
    QAction*    addAct;
    QAction*    saveAct;
    QAction*    closeAct;
    QAction*    UserPreferAct;
    QAction*    UserExitAct;

    // edit menu
    QAction*    transAct;
    //QAction *insertAct;
    QAction*    insertVertexAct;
    QAction*    insertSegmentAct;
    QAction*    insertRectAct;
    QAction*    insertArcAct;
    //QAction *selectAct;
    QAction*    assignVertexAct;
    QAction*    assignVertexWeightAct;
    QAction*    assignSegmentAct;
    QAction*    assignSubdomainAct;
    QAction*    setAttributeAct;
    QAction*    deleteAct;
    QAction*    scaleAct;

    // view menu
    QAction*    viewVerticesAct;
    QAction*    viewWeightsAct;
    QAction*    viewVertexLabelsAct;
    QAction*    viewSegmentsAct;
    QAction*    viewSegmentLabelsAct;
    QAction*    viewSubdomainsAct;
    QAction*    viewConvexHullAct;
    QAction*    viewTrianglesAct;
    QAction*    viewMeshVertexLabelsAct;
    QAction*    viewTriangleInfoAct;
    QAction*    viewOthocirclesAct;
    QAction*    viewOthocircleAct;
    QAction*    viewOthocenterAct;
    QAction*    viewPowerDiagramAct;
    QAction*    viewMassCentersAct;
    QAction*    viewLaplacianCentersAct;
    QAction*    viewPowerCellAct;
    QAction*    viewMassCenterAct;
    QAction*    viewBboxAct;
    QAction*    viewBgMeshAct;

    QAction*    MoveLeftAct;
    QAction*    MoveRightAct;
    QAction*    MoveUpAct;
    QAction*    MoveDownAct;
    QAction*    zoomInAct;
    QAction*    zoomOutAct;
    QAction*    zoomSelectAct;
    QAction*    viewResetAct;

    // mesh Menu
    QAction*    detri2OptionsAct;
    QAction*    weightedDelaunayAct;
    QAction*    furthestDelaunayAct;
    QAction*    detri2HDEAct;
    QAction*    detri2SortAct;
    QAction*    cleanMeshAct;

    void createActions();
    void createMenus();

protected:
    void mousePressEvent(QMouseEvent *event) override;
    void mouseMoveEvent(QMouseEvent *event) override;
    void mouseReleaseEvent(QMouseEvent *event) override;
    void wheelEvent(QWheelEvent *e) override;
    //void keyPressEvent(QKeyEvent *event) override;

    void paintEvent(QPaintEvent *event) override;
    void resizeEvent(QResizeEvent *event) override;

public:
    int  set_active_vertex(detri2::Vertex *thevertex);
    int  is_vertex_active(detri2::Vertex *thevertex);
    int  remove_active_vertex(detri2::Vertex *thevertex);
    int  add_active_vertex(detri2::Vertex *thevertex);
    int  set_active_segment(detri2::Triang *theseg);
    int  is_segment_active(detri2::Triang *theseg);
    int  remove_active_segment(detri2::Triang *theseg);
    int  add_active_segment(detri2::Triang *theseg);
    int  set_active_trilist(detri2::arraypool *newtrilist);

    void calc_trans(double xmin, double xmax, double ymin, double ymax);
    void get_selection_region(double& xmin, double& xmax, double& ymin, double& ymax);
    void draw_scene(QPainter &painter);
};

#endif
