#include "meshoptionpanel.h"
#include "scribblearea.h"
#include <QtWidgets>
#include <QFileDialog>
#include <QDir>
#include <QDebug>


meshoptionPanel::meshoptionPanel(void *scribblearea, QWidget *parent) : QWidget(parent)
{
    _scribblearea = scribblearea;

    setWindowTitle(tr("Mesh algorithms & options"));

    QVBoxLayout *verticalLayout;
    verticalLayout = new QVBoxLayout(this);

    QGroupBox *groupBox_Algorithms = new QGroupBox(tr("Mesh algorithms"));
    op_Convexhull = new QRadioButton("Convex hull (point sets)");
    op_Delaunay = new QRadioButton("(Delaunay) Triangulations (point sets)");
    op_CDT      = new QRadioButton("Constrained (Delaunay) Triangulations (line segments)");
    op_Quality  = new QRadioButton("Quality mesh generation (polygonal regions)");
    op_Flip     = new QRadioButton("Lawson flipping (using Delaunay criterion)");
    op_Smooth   = new QRadioButton("Smoth mesh vertices");
    op_Delaunay->setChecked(true);
    QVBoxLayout *vbox_Algo = new QVBoxLayout;
    vbox_Algo->addWidget(op_Convexhull);
    vbox_Algo->addWidget(op_Delaunay);
    vbox_Algo->addWidget(op_CDT);
    vbox_Algo->addWidget(op_Quality);
    vbox_Algo->addWidget(op_Flip);
    vbox_Algo->addWidget(op_Smooth);
    vbox_Algo->addStretch(1);
    groupBox_Algorithms->setLayout(vbox_Algo);
    verticalLayout->addWidget(groupBox_Algorithms);

    /////////////////////////////////////////////////////////////

    // Delaunay options
    op_DT  = new QRadioButton(tr("nearest-point"));
    op_FDT = new QRadioButton(tr("furthest-point"));
    op_DT->setChecked(true); // default
    op_weighted = new QCheckBox("Weighted");
    op_weighted->setChecked(true); // default
    op_no_sort = new QCheckBox("no sort");
    op_no_flip = new QCheckBox("no edge flip (Delaunay)");

    // Quality options
    QLabel *op_maxarea_label = new QLabel(tr("Maxmimum triangle area (> 0)"));
    line_op_maxarea = new QLineEdit;
    line_op_maxarea->setText(tr("0"));
    QLabel *op_minangle_label = new QLabel(tr("Minimum triangle angle (degree)"));
    line_op_minangle = new QLineEdit;
    line_op_minangle->setText(tr("20"));
    QLabel *op_minlen_label = new QLabel(tr("Minimum edge length (> 0)"));
    line_op_minlen = new QLineEdit;
    line_op_minlen->setText(tr("0"));
    op_metric = new QCheckBox(tr("Use mesh size (defined on vertices, segments, subregions)"));

    // Smooth options
    QLabel *smooth_methods_label = new QLabel(tr("Smooth method"));
    _cb_smooth_methods = new QComboBox;
    _cb_smooth_methods->addItem(tr("Laplacian"));
    _cb_smooth_methods->addItem(tr("CVT"));
    QLabel *smooth_iter_label = new QLabel(tr("Number of iterations"));
    line_smooth_iters = new QLineEdit();
    line_smooth_iters->setText(tr("1")); // default is 1 iteration

    // File dump options
    op_dump_to_ucd = new QCheckBox(tr("Save intermediate meshes"));
    op_dump_lift_map = new QCheckBox(tr("Save lifting maps"));
    line_dump_file_name = new QLineEdit;
    line_dump_file_name->setText(QDir::currentPath() + "/sol");
    btnfilename = new QPushButton(tr("file name"));

    QToolBox *toolbox = new QToolBox;
    verticalLayout->addWidget(toolbox);

    QWidget *page = new QWidget;
    QGridLayout *layout = new QGridLayout(page);
    //layout->setColumnStretch(1, 1);
    //layout->setColumnMinimumWidth(1, 250);
    layout->addWidget(op_DT, 0, 0, 1, 1);
    layout->addWidget(op_FDT, 0, 1, 1, 1);
    layout->addWidget(op_weighted, 0,2, 1, 1);
    layout->addWidget(op_no_sort, 1,0,1,1);
    layout->addWidget(op_no_flip, 1,1,1,2);
    toolbox->addItem(page, tr("Incremental (Delaunay) triangulation options"));

    page = new QWidget;
    layout = new QGridLayout(page);
    //layout->setColumnStretch(1, 1);
    //layout->setColumnMinimumWidth(1, 250);
    //layout->addWidget(groupBox_Quality, 0, 0);
    layout->addWidget(op_minangle_label, 0, 0, 1, 1);
    layout->addWidget(line_op_minangle, 0, 1, 1, 1);
    layout->addWidget(op_maxarea_label, 1, 0, 1, 1);
    layout->addWidget(line_op_maxarea, 1, 1, 1, 1);
    layout->addWidget(op_minlen_label, 2, 0, 1, 1);
    layout->addWidget(line_op_minlen, 2, 1, 1, 1);
    //layout->addWidget(op_verbose_label, 2, 0, 1, 1);
    //layout->addWidget(line_op_verbose, 2, 1, 1, 1);
    layout->addWidget(op_metric, 3, 0, 1, 2);
    toolbox->addItem(page, tr("Mesh quality options"));

    page = new QWidget;
    layout = new QGridLayout(page);
    //layout->setColumnStretch(1, 1);
    //layout->setColumnMinimumWidth(1, 250);
    //layout->addWidget(groupBox_Smooth, 0, 0);
    layout->addWidget(smooth_methods_label, 0, 0);
    layout->addWidget(_cb_smooth_methods, 0, 1);
    layout->addWidget(smooth_iter_label, 1, 0);
    layout->addWidget(line_smooth_iters, 1, 1);
    toolbox->addItem(page, tr("Smooth options"));

    page = new QWidget;
    layout = new QGridLayout(page);
    //layout->setColumnStretch(1, 1);
    //layout->setColumnMinimumWidth(1, 250);
    //layout->addWidget(groupBox_dump, 0, 0);
    layout->addWidget(op_dump_to_ucd, 0,0,1,1);
    layout->addWidget(op_dump_lift_map, 0,1,1,1);
    layout->addWidget(btnfilename, 0,2,1,1);
    layout->addWidget(line_dump_file_name, 1,0,1,3);
    toolbox->addItem(page, tr("Save intermediate meshes options"));

    /////////////////////////////////////////////////////////////

    btnGenerate = new QPushButton;
    btnGenerate->setText(tr("Generate"));
    btnAnimation = new QPushButton;
    btnAnimation->setText(tr("Show Animation"));
    btnClear = new QPushButton;
    btnClear->setText(tr("Clear"));
    btnClose = new QPushButton;
    btnClose->setText(tr("Close"));
    QGridLayout *hbox_Btn = new QGridLayout;
    hbox_Btn->addWidget(btnGenerate, 0,0);
    hbox_Btn->addWidget(btnAnimation,0,1);
    hbox_Btn->addWidget(btnClear,1,0);
    hbox_Btn->addWidget(btnClose,1,1);
    QGroupBox *groupBox_Btn = new QGroupBox;
    groupBox_Btn->setLayout(hbox_Btn);
    verticalLayout->addWidget(groupBox_Btn);

    connect(btnfilename, SIGNAL(clicked()), this, SLOT(on_filename()));

    connect(btnGenerate, SIGNAL(clicked()), this, SLOT(generate()));
    connect(btnAnimation, SIGNAL(clicked()), this, SLOT(show_animation()));
    connect(btnClear, SIGNAL(clicked()), this, SLOT(clear_mesh()));
    connect(btnClose, SIGNAL(clicked()), this, SLOT(close()));
}

void meshoptionPanel::on_filename()
{
   QString filepath = QDir::currentPath(); // get current working directory
   if (_scribblearea != NULL) {
     filepath = ((ScribbleArea *)_scribblearea)->_file_last_used_path;
   }

    QString fileName = QFileDialog::getSaveFileName(this,
                                              tr("Dump file name"),
                                              filepath
                                              );
    if (!fileName.isEmpty()) {
      line_dump_file_name->setText(fileName);
    }
}

void meshoptionPanel::generate()
{
    qDebug()<<"Generate btn clicked";
    if (_scribblearea != NULL) {
      ((ScribbleArea *)_scribblearea)->generate_mesh_act();
    }
}

void meshoptionPanel::show_animation()
{
    qDebug()<<"Show animation btn clicked";
    if (_scribblearea != NULL) {
      ((ScribbleArea *)_scribblearea)->generate_mesh_ani_act();
    }
}

void meshoptionPanel::clear_mesh()
{
    if (_scribblearea != NULL) {
      ((ScribbleArea *)_scribblearea)->clean_mesh_act();
    }
}

void meshoptionPanel::close()
{
    //qDebug()<<"close btn clicked";
    hide(); //close();
}
