#include "newfilepanel.h"
#include "vertexpanel.h"
#include "scribblearea.h"

#include <QDialogButtonBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QMessageBox>
#include <QDebug>

GeometryInputPanel::GeometryInputPanel(int geomtype, QDialog *parent) : QDialog(parent)
{
    setWindowTitle(tr("Geometry Parameters"));

    lbl_centx = new QLabel("center x");
    lbl_centy = new QLabel("center y");
    lbl_rect_width = new QLabel("rect width");
    lbl_rect_heigh = new QLabel("rect height");
    QLabel *lbl_cir_radius = new QLabel("circle radius");
    QLabel *lbl_cir_start = new QLabel("start angle (degree)");
    QLabel *lbl_cir_end = new QLabel("end angle (degree)");
    QLabel *lbl_cir_segnum = new QLabel("# segments");
    QLabel *lbl_seg_tag = new QLabel("boundary tag");
    centx = new QLineEdit();
    centy = new QLineEdit();
    rect_width = new QLineEdit();
    rect_height = new QLineEdit();
    cir_radius = new QLineEdit();
    cir_angle_start = new QLineEdit();
    cir_angle_end = new QLineEdit();
    cir_segnum = new QLineEdit();
    segment_tag = new QLineEdit();

    // Set the default values
    centx->setText(tr("0"));
    centy->setText(tr("0"));
    rect_width->setText(tr("2.0"));
    rect_height->setText(tr("2.0"));
    cir_radius->setText(tr("1.0"));
    cir_angle_start->setText(tr("0"));
    cir_angle_end->setText(tr("360"));
    cir_segnum->setText(tr("18"));
    segment_tag->setText(tr("-1"));

    QGridLayout *rectLayout = new QGridLayout;
    rectLayout->addWidget(lbl_centx, 0,0);
    rectLayout->addWidget(lbl_centy, 1,0);
    rectLayout->addWidget(centx,0,1);
    rectLayout->addWidget(centy,1,1);
    if (geomtype == 1) {
      rectLayout->addWidget(lbl_rect_width, 2,0);
      rectLayout->addWidget(lbl_rect_heigh, 3,0);
      rectLayout->addWidget(rect_width, 2,1);
      rectLayout->addWidget(rect_height, 3,1);
      rectLayout->addWidget(lbl_seg_tag, 4,0);
      rectLayout->addWidget(segment_tag, 4,1);
    } else if (geomtype == 2) {
      rectLayout->addWidget(lbl_cir_radius, 2,0);
      rectLayout->addWidget(lbl_cir_start, 3,0);
      rectLayout->addWidget(lbl_cir_end, 4,0);
      rectLayout->addWidget(lbl_cir_segnum, 5,0);
      rectLayout->addWidget(cir_radius, 2,1);
      rectLayout->addWidget(cir_angle_start, 3,1);
      rectLayout->addWidget(cir_angle_end, 4,1);
      rectLayout->addWidget(cir_segnum, 5,1);
      rectLayout->addWidget(lbl_seg_tag, 6,0);
      rectLayout->addWidget(segment_tag, 6,1);
    }
    //rectLayout->setHorizontalSpacing(10);
    //rectLayout->setVerticalSpacing(10);
    QGroupBox *groupBox_rect = new QGroupBox;
    groupBox_rect->setLayout(rectLayout);

    QLabel *lbl_rot_cx = new QLabel("Rotate center x");
    QLabel *lbl_rot_cy = new QLabel("Rotate center y");
    QLabel *lbl_rot_angle = new QLabel("Rotate angle (degree)");
    QLabel *lbl_trans_x = new QLabel("Translate dx");
    QLabel *lbl_trans_y = new QLabel("Translate dy");
    rot_cx = new QLineEdit();
    rot_cy = new QLineEdit();
    rot_angle = new QLineEdit();
    trans_x = new QLineEdit();
    trans_y = new QLineEdit();
    use_trans = new QCheckBox(tr("Use Translation and Rotation"));
    QGridLayout *transLayout = new QGridLayout;
    transLayout->addWidget(use_trans, 0,0);
    transLayout->addWidget(lbl_trans_x, 1,0);
    transLayout->addWidget(lbl_trans_y, 2,0);
    transLayout->addWidget(trans_x, 1,1);
    transLayout->addWidget(trans_y, 2,1);
    transLayout->addWidget(lbl_rot_cx, 3,0);
    transLayout->addWidget(lbl_rot_cy, 4,0);
    transLayout->addWidget(rot_cx, 3,1);
    transLayout->addWidget(rot_cy, 4,1);
    transLayout->addWidget(lbl_rot_angle, 5,0);
    transLayout->addWidget(rot_angle, 5,1);
    //transLayout->setHorizontalSpacing(10);
    //transLayout->setVerticalSpacing(10);
    //QGroupBox *groupBox_trans = new QGroupBox(tr("Translation and Rotation"));

    QGroupBox *groupBox_trans = new QGroupBox();
    groupBox_trans->setLayout(transLayout);
    //extension->setLayout(transLayout);

    trans_x->setText(tr("0"));
    trans_y->setText(tr("0"));
    rot_cx->setText(tr("0"));
    rot_cy->setText(tr("0"));
    rot_angle->setText(tr("0"));

    QDialogButtonBox* bb = new QDialogButtonBox(
      QDialogButtonBox::Ok | QDialogButtonBox::Cancel);

    QPushButton *btnOk = bb->button(QDialogButtonBox::Ok);
    btnOk->setAutoDefault(true);
    btnOk->setDefault(true);

    QPushButton *btnCancel = bb->button(QDialogButtonBox::Cancel);
    btnCancel->setAutoDefault(false);
    btnCancel->setDefault(false);

    moreButton = new QPushButton(tr("&More"));
    moreButton->setCheckable(true);
    moreButton->setAutoDefault(false);

    bb->addButton(moreButton, QDialogButtonBox::ActionRole);

    connect(btnOk, SIGNAL(clicked()), this, SLOT(accept()));
    connect(btnCancel, SIGNAL(clicked()), this, SLOT(reject()));

    connect(moreButton, &QAbstractButton::toggled, groupBox_trans, &QWidget::setVisible);

    QGridLayout *mainLayout = new QGridLayout;
    mainLayout->setSizeConstraint(QLayout::SetFixedSize);
    mainLayout->addWidget(groupBox_rect, 0, 0);
    mainLayout->addWidget(groupBox_trans, 1, 0);
    mainLayout->addWidget(bb, 2, 0);

    groupBox_trans->hide(); // default do not show

    setLayout(mainLayout);
}

//============================================================================

newfilePanel::newfilePanel(void *scribblearea, int new_or_add, QWidget *parent) : QWidget(parent)
{
    _scribblearea = scribblearea;
    _new_or_add = new_or_add;

    if (_new_or_add == 0) {
      setWindowTitle(tr("New geometry options"));
    } else {
      setWindowTitle(tr("Add geometry options"));
    }

    op_empty  = new QRadioButton(tr("empty geometry"));
    op_vertex = new QRadioButton(tr("single vertex"));
    op_rbox   = new QRadioButton(tr("(random) point set"));
    op_rect   = new QRadioButton(tr("rectangle"));
    op_circle = new QRadioButton(tr("arc (circle or n-gon)"));
    op_rbox->setChecked(true); // default

    QVBoxLayout *vbox = new QVBoxLayout;
    if (_new_or_add == 0) {
      vbox->addWidget(op_empty);
    } else {
      vbox->addWidget(op_vertex);
    }
    vbox->addWidget(op_rbox);
    vbox->addWidget(op_rect);
    vbox->addWidget(op_circle);
    vbox->addStretch(1);
    QGroupBox *groupBox = new QGroupBox();
    groupBox->setLayout(vbox);

    QPushButton *openFileNamesButton =
            new QPushButton(tr("from file..."));

    btnNew = new QPushButton(tr("New"));
    btnAdd = new QPushButton(tr("Add"));
    btnCancel = new QPushButton(tr("Cancel"));

    QDialogButtonBox *buttonBox = new QDialogButtonBox(Qt::Vertical);
    if (_new_or_add == 0) {
      buttonBox->addButton(btnNew, QDialogButtonBox::ActionRole);
      btnNew->setDefault(true);
    } else {
      buttonBox->addButton(btnAdd, QDialogButtonBox::ActionRole);
      btnAdd->setDefault(true);
    }
    buttonBox->addButton(btnCancel, QDialogButtonBox::ActionRole);

    QGridLayout *mainLayout = new QGridLayout;
    mainLayout->setSizeConstraint(QLayout::SetFixedSize);
    //mainLayout->addLayout(leftLayout, 0, 0);
    mainLayout->addWidget(groupBox, 0, 0);
    if (_new_or_add > 0) {
      mainLayout->addWidget(openFileNamesButton, 1,0);
    }
    mainLayout->addWidget(buttonBox, 0, 1);

    connect(btnNew, SIGNAL(clicked()), this, SLOT(on_new()));
    connect(btnAdd, SIGNAL(clicked()), this, SLOT(on_add()));
    connect(btnCancel, SIGNAL(clicked()), this, SLOT(on_cancel()));

    connect(openFileNamesButton, SIGNAL(clicked()), this, SLOT(get_from_File()));

    setLayout(mainLayout);
}

void newfilePanel::on_new()
{
    if (op_empty->isChecked()) {
      // Clear the old geomtry and return an empty geometry.
      ScribbleArea *SA = (ScribbleArea *) _scribblearea;
      // Reset everything
      SA->clean_scribble();
      SA->reset_scribble();
      // Initialize an empty geometry
      detri2_init_with_bbox(&SA->_geom, -1., 1., -1., 1.);
      SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                   SA->_geom->io_ymin, SA->_geom->io_ymax);
      SA->update();
    } else if (op_rbox->isChecked()) {
      bool ok;
      QString cmd = QInputDialog::getText(this, tr("Random point set"),
                                           tr("number of points"), QLineEdit::Normal,
                                           tr("50"), // QDir::home().dirName()
                                           &ok);
      if (ok && !cmd.isEmpty()) {
        // Parse commands
        QStringList ss = cmd.split(" ");
        int argc = ss.size()+2;
        qDebug()<<" argc="<<argc;
        char **argv = new char*[argc];
        for (int i = 0; i < argc; i++) {
          argv[i] = new char[100];
        }
        strcpy(argv[0], "rbox");
        strcpy(argv[1], "D2");
        qDebug()<<" command line: "<<argv[0]<<" "<<argv[1];
        for (int i = 0; i < ss.size(); i++) {
          QByteArray ba = ss[i].toLatin1();
          strcpy(argv[i+2], ba.data());
          qDebug()<<" "<<argv[i+2];
        }

        ScribbleArea *SA = (ScribbleArea *) _scribblearea;
        // Reset everything
        SA->clean_scribble();
        SA->reset_scribble();

        detri2_rbox(&SA->_geom, argc, argv);

        for (int i = 0; i < argc; i++) delete [] argv[i];
        delete [] argv;

        SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                       SA->_geom->io_ymin, SA->_geom->io_ymax);
        //SA->_object_valid=0;
        SA->update();
      }
    } else if (op_rect->isChecked()) {
      GeometryInputPanel *Panel = new GeometryInputPanel(1);
      if(Panel->exec() == QDialog::Accepted){
        // You can access everything you need in dialog object
        //qDebug()<<"You can access everything you need in dialog object from here";
        double ctx = Panel->centx->text().toDouble();
        double cty = Panel->centy->text().toDouble();
        double wid = Panel->rect_width->text().toDouble();
        double hei = Panel->rect_height->text().toDouble();
        int tag = Panel->segment_tag->text().toInt();
        qDebug()<<"Your parameters";
        qDebug()<<"  center x = "<<ctx;
        qDebug()<<"  center y = "<<cty;
        qDebug()<<"  rect  width = "<<wid;
        qDebug()<<"  rect height = "<<hei;
        qDebug()<<"          tag = "<<tag;
        if ((wid > 0) && (hei > 0)) {
          // Clear everything.
          ScribbleArea *SA = (ScribbleArea *) _scribblearea;
          // Reset everything
          SA->clean_scribble();
          SA->reset_scribble();
          // Create a new rectangle
          if (Panel->use_trans->isChecked()) {
            // Translate the rectangle.
            double rcx = Panel->rot_cx->text().toDouble();
            double rcy = Panel->rot_cy->text().toDouble();
            double ang = Panel->rot_angle->text().toDouble();
            double dx  = Panel->trans_x->text().toDouble();
            double dy  = Panel->trans_y->text().toDouble();
            detri2_add_rectangle(&SA->_geom, ctx, cty, wid, hei, tag,
                                 true, rcx, rcy, ang, dx, dy);
          } else {
            detri2_add_rectangle(&SA->_geom, ctx, cty, wid, hei, tag,
                                 false, 0, 0, 0, 0, 0);
          }
          SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                         SA->_geom->io_ymin, SA->_geom->io_ymax);
          //SA->_object_valid=0;
          SA->update();
        } else {
          QMessageBox msgBox;
          msgBox.setText(tr("width or height must not be 0"));
          msgBox.exec();
        }
      }
      delete Panel;
    } else if (op_circle->isChecked()) {
      GeometryInputPanel *Panel = new GeometryInputPanel(2);
      if(Panel->exec() == QDialog::Accepted){
        // You can access everything you need in dialog object
        //qDebug()<<"You can access everything you need in dialog object from here";
        double rcx = Panel->centx->text().toDouble();
        double rcy = Panel->centy->text().toDouble();
        double radius = Panel->cir_radius->text().toDouble();
        double ang1 = Panel->cir_angle_start->text().toDouble();
        double ang2 = Panel->cir_angle_end->text().toDouble();
        double segnum = Panel->cir_segnum->text().toDouble();
        int tag = Panel->segment_tag->text().toInt();
        qDebug()<<"Your parameters";
        qDebug()<<"  center x = "<<rcx;
        qDebug()<<"  center y = "<<rcy;
        qDebug()<<"  radius = "<<radius;
        qDebug()<<"  start angle = "<<ang1;
        qDebug()<<"    end angle = "<<ang2;
        qDebug()<<"       segnum = "<<segnum;
        qDebug()<<"          tag = "<<tag;
        if ((radius > 0) && (ang2 > ang1) && (segnum > 1)) {
          // Clear everything.
          ScribbleArea *SA = (ScribbleArea *) _scribblearea;
          // Reset everything
          SA->clean_scribble();
          SA->reset_scribble();
          // Create a new rectangle
          if (Panel->use_trans->isChecked()) {
            // Translate the rectangle.
            double rcx = Panel->rot_cx->text().toDouble();
            double rcy = Panel->rot_cy->text().toDouble();
            double ang = Panel->rot_angle->text().toDouble();
            double tx  = Panel->trans_x->text().toDouble();
            double ty  = Panel->trans_y->text().toDouble();
            detri2_add_circle(&SA->_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                 true, rcx, rcy, ang, tx, ty);
          } else {
            detri2_add_circle(&SA->_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                 false, 0, 0, 0, 0, 0);
          }
          SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                         SA->_geom->io_ymin, SA->_geom->io_ymax);
          //SA->_object_valid=0;
          SA->update();
        }
      }
      delete Panel;
    }

    hide();
}

void newfilePanel::on_add()
{
    if (op_empty->isChecked()) {
      //assert(0); // should not be called
      /*
      // Clear the old geomtry and return an empty geometry.
      ScribbleArea *SA = (ScribbleArea *) _scribblearea;
      // Reset everything
      SA->clean_scribble();
      SA->reset_scribble();
      // Initialize an empty geometry
      detri2_init_with_bbox(&SA->_geom, -1., 1., -1., 1.);
      SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                   SA->_geom->io_ymin, SA->_geom->io_ymax);
      SA->update();
      */
    }

    if (op_vertex->isChecked()) {
      ScribbleArea *SA = (ScribbleArea *) _scribblearea;

      if (SA->_vertexPanel == NULL) {
        SA->_vertexPanel = new vertexPanel(this); // only initialise it once.
      }
      if (SA->_active_vertex) {
        // Disable selected vertices.
        SA->set_active_vertex(NULL);
      }
      // Dafault at the center location of the bbox.
      double ocx = 0.5 * (SA->_geom->io_xmin + SA->_geom->io_xmax);
      double ocy = 0.5 * (SA->_geom->io_ymin + SA->_geom->io_ymax);
      SA->_vertexPanel->_idx->setText(QString("index = 0"));
      SA->_vertexPanel->_x->setText(QString::number(ocx));
      SA->_vertexPanel->_y->setText(QString::number(ocy));
      SA->_vertexPanel->_w->setText(QString::number(0.0));
      SA->_vertexPanel->_val->setText(QString::number(0.0));

      SA->_vertexPanel->show();
    } else if (op_rbox->isChecked()) {
      bool ok;
      QString cmd = QInputDialog::getText(this, tr("Random point set"),
                                           tr("number of points"), QLineEdit::Normal,
                                           tr("50"), // QDir::home().dirName()
                                           &ok);
      if (ok && !cmd.isEmpty()) {
        // Parse commands
        QStringList ss = cmd.split(" ");
        int argc = ss.size()+2;
        qDebug()<<" argc="<<argc;
        char **argv = new char*[argc];
        for (int i = 0; i < argc; i++) {
          argv[i] = new char[100];
        }
        strcpy(argv[0], "rbox");
        strcpy(argv[1], "D2");
        qDebug()<<" command line: "<<argv[0]<<" "<<argv[1];
        for (int i = 0; i < ss.size(); i++) {
          QByteArray ba = ss[i].toLatin1();
          strcpy(argv[i+2], ba.data());
          qDebug()<<" "<<argv[i+2];
        }

        ScribbleArea *SA = (ScribbleArea *) _scribblearea;

        // Reset everything
        //SA->clean_scribble();
        //SA->reset_scribble();

        detri2_rbox(&SA->_geom, argc, argv);

        for (int i = 0; i < argc; i++) delete [] argv[i];
        delete [] argv;

        SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                       SA->_geom->io_ymin, SA->_geom->io_ymax);
        SA->_object_valid=0;
        SA->update();
      }
    } else if (op_rect->isChecked()) {
      GeometryInputPanel *Panel = new GeometryInputPanel(1);
      if(Panel->exec() == QDialog::Accepted){
        // You can access everything you need in dialog object
        //qDebug()<<"You can access everything you need in dialog object from here";
        double ctx = Panel->centx->text().toDouble();
        double cty = Panel->centy->text().toDouble();
        double wid = Panel->rect_width->text().toDouble();
        double hei = Panel->rect_height->text().toDouble();
        int tag = Panel->segment_tag->text().toInt();
        qDebug()<<"Your parameters";
        qDebug()<<"  center x = "<<ctx;
        qDebug()<<"  center y = "<<cty;
        qDebug()<<"  rect  width = "<<wid;
        qDebug()<<"  rect height = "<<hei;
        qDebug()<<"          tag = "<<tag;
        if ((wid > 0) && (hei > 0)) {
          // Clear everything.
          ScribbleArea *SA = (ScribbleArea *) _scribblearea;
          // Reset everything
          //SA->clean_scribble();
          //SA->reset_scribble();
          // Create a new rectangle
          if (Panel->use_trans->isChecked()) {
            // Translate the rectangle.
            double rcx = Panel->rot_cx->text().toDouble();
            double rcy = Panel->rot_cy->text().toDouble();
            double ang = Panel->rot_angle->text().toDouble();
            double dx  = Panel->trans_x->text().toDouble();
            double dy  = Panel->trans_y->text().toDouble();
            detri2_add_rectangle(&SA->_geom, ctx, cty, wid, hei, tag,
                                 true, rcx, rcy, ang, dx, dy);
          } else {
            detri2_add_rectangle(&SA->_geom, ctx, cty, wid, hei, tag,
                                 false, 0, 0, 0, 0, 0);
          }
          SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                         SA->_geom->io_ymin, SA->_geom->io_ymax);
          SA->_object_valid=0;
          SA->update();
        } else {
          QMessageBox msgBox;
          msgBox.setText(tr("width or height must not be 0"));
          msgBox.exec();
        }
      }
      delete Panel;
    } else if (op_circle->isChecked()) {
      GeometryInputPanel *Panel = new GeometryInputPanel(2);
      if(Panel->exec() == QDialog::Accepted){
        // You can access everything you need in dialog object
        //qDebug()<<"You can access everything you need in dialog object from here";
        double rcx = Panel->centx->text().toDouble();
        double rcy = Panel->centy->text().toDouble();
        double radius = Panel->cir_radius->text().toDouble();
        double ang1 = Panel->cir_angle_start->text().toDouble();
        double ang2 = Panel->cir_angle_end->text().toDouble();
        double segnum = Panel->cir_segnum->text().toDouble();
        int tag = Panel->segment_tag->text().toInt();
        qDebug()<<"Your parameters";
        qDebug()<<"  center x = "<<rcx;
        qDebug()<<"  center y = "<<rcy;
        qDebug()<<"  radius = "<<radius;
        qDebug()<<"  start angle = "<<ang1;
        qDebug()<<"    end angle = "<<ang2;
        qDebug()<<"       segnum = "<<segnum;
        qDebug()<<"          tag = "<<tag;
        if ((radius > 0) && (ang2 > ang1) && (segnum > 1)) {
          ScribbleArea *SA = (ScribbleArea *) _scribblearea;
          // Reset everything
          //SA->clean_scribble();
          //SA->reset_scribble();
          // Create a new rectangle
          if (Panel->use_trans->isChecked()) {
            // Translate the rectangle.
            double rcx = Panel->rot_cx->text().toDouble();
            double rcy = Panel->rot_cy->text().toDouble();
            double ang = Panel->rot_angle->text().toDouble();
            double tx  = Panel->trans_x->text().toDouble();
            double ty  = Panel->trans_y->text().toDouble();
            detri2_add_circle(&SA->_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                 true, rcx, rcy, ang, tx, ty);
          } else {
            detri2_add_circle(&SA->_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                 false, 0, 0, 0, 0, 0);
          }
          SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                         SA->_geom->io_ymin, SA->_geom->io_ymax);
          SA->_object_valid=0;
          SA->update();
        }
      }
      delete Panel;
    }

    hide();
}

void newfilePanel::on_cancel()
{
    hide();
}

void newfilePanel::get_from_File()
{
    //const QFileDialog::Options options = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;
    QString fileName = QFileDialog::getOpenFileName(this,
                                tr("Add geometry from file"),
                                "",
                                tr("*.node;;*.poly;;*.smesh;;*.mesh"),
                                &selectedFilter);
    if (!fileName.isEmpty()) {
       // openFileNameLabel->setText(fileName);
       qDebug()<<fileName;
       QByteArray ba = fileName.toLatin1();
       const char *c_str2 = ba.data();

       int argc = 2;
       char tmp_str1[256], tmp_str2[256];
       strcpy(tmp_str1, "dummy");
       strcpy(tmp_str2, c_str2);
       char *argv[2];
       argv[0] = tmp_str1;
       argv[1] = tmp_str2;

       detri2::Triangulation *Tr = new detri2::Triangulation();
       if (Tr->parse_commands(argc, argv)) {
         if (Tr->read_mesh()) {
           qDebug()<<"Call detri2_add_geometry()";
           int np = Tr->ct_in_vrts;
           double *ptlist = new double[np * 2];
           for (int i = 0; i < np; i++) {
             detri2::Vertex *v = &(Tr->in_vrts[i]);
             ptlist[i*2] = v->crd[0];
             ptlist[i*2+1] = v->crd[1];
           }
           int ne = Tr->tr_segs->objects;
           int *elist = NULL;
           int *etaglist = NULL;
           if (ne > 0) {
             elist = new int[ne * 2];
             etaglist = new int[ne];
             for (int i = 0; i < ne; i++) {
               detri2::Triang *seg = (detri2::Triang *) Tr->tr_segs->get(i);
               elist[i*2] = seg->vrt[0]->idx - Tr->io_firstindex;
               elist[i*2+1] = seg->vrt[1]->idx - Tr->io_firstindex;
               etaglist[i] = seg->tag;
             }
           } // ne > 0

           ScribbleArea *SA = (ScribbleArea *) _scribblearea;
           detri2_add_geometry(&(SA->_geom), np, ptlist, ne, elist, etaglist);

           SA->calc_trans(SA->_geom->io_xmin, SA->_geom->io_xmax,
                          SA->_geom->io_ymin, SA->_geom->io_ymax);
           SA->_object_valid=0;
           SA->update();

           delete [] ptlist;
           if (ne > 0){
             delete [] elist;
             delete [] etaglist;
           }
         } // if (Tr->read_mesh())
       } // if (Tr->parse_commands(argc, argv))

       delete Tr;
       hide();
    }
}
